//! Async [Monad] based on [`Pin<Box<dyn Future>>`] (see: [`Pin`], [`Box::pin`], [`Future`]).
//! This generally allows just using `.await` on wrapped instances.
//!
//! For sync, see [`super::solo`].

use std::{future::Future, pin::Pin};

use futures::{
    future::{select, Either, Shared},
    join, FutureExt,
};

use crate::func::*;

#[derive(CovariantFunctor)]
pub struct FutureClass;

impl WeakFunctor for FutureClass {
    type F<'a, A: 'a> = Pin<Box<dyn 'a + Future<Output = A>>>;
}

impl Functor for FutureClass {
    fn fmap<'a, A: 'a, B: 'a>(f: impl 'a + FnOnce(A) -> B, fa: Self::F<'a, A>) -> Self::F<'a, B> {
        Box::pin(async { f(fa.await) })
    }

    fn replace<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, b: B) -> Self::F<'a, B> {
        Box::pin(async {
            fa.await;
            b
        })
    }
}

impl Pure for FutureClass {
    fn pure<'a, A: 'a>(a: A) -> Self::F<'a, A> {
        Box::pin(async { a })
    }
}

impl ApplicativeSeq for FutureClass {
    fn seq<'a, A: 'a, B: 'a>(
        ff: Self::F<'a, impl 'a + FnOnce(A) -> B>,
        fa: Self::F<'a, A>,
    ) -> Self::F<'a, B> {
        Box::pin(async {
            let (f, a) = join!(ff, fa);
            f(a)
        })
    }
}

impl ApplicativeLA2 for FutureClass {
    fn la2<'a, A: 'a, B: 'a, C: 'a>(
        f: impl 'a + FnOnce(A, B) -> C,
        fa: Self::F<'a, A>,
        fb: Self::F<'a, B>,
    ) -> Self::F<'a, C> {
        Box::pin(async {
            let (a, b) = join!(fa, fb);
            f(a, b)
        })
    }
}

impl ApplicativeTuple for FutureClass {
    fn tuple<'a, A: 'a, B: 'a>((fa, fb): (Self::F<'a, A>, Self::F<'a, B>)) -> Self::F<'a, (A, B)>
    where
        Self: 'a,
    {
        Box::pin(async { join!(fa, fb) })
    }
}

impl ApplicativeSelect for FutureClass {
    fn select<'a, A: 'a, B: 'a, C: 'a>(
        f: impl 'a + FnOnce(Selected<'a, A, B, Self>) -> C,
        fa: Self::F<'a, A>,
        fb: Self::F<'a, B>,
    ) -> Self::F<'a, C>
    where
        Self: 'a,
    {
        Box::pin(async {
            match select(fa, fb).await {
                Either::Left((a, fb)) => f(Selected::A(a, fb)),
                Either::Right((b, fa)) => f(Selected::B(fa, b)),
            }
        })
    }
}

impl Applicative for FutureClass {
    fn discard_first<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, fb: Self::F<'a, B>) -> Self::F<'a, B> {
        Box::pin(async { join!(fa, fb).1 })
    }

    fn discard_second<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, fb: Self::F<'a, B>) -> Self::F<'a, A> {
        Box::pin(async { join!(fa, fb).0 })
    }
}

impl Monad for FutureClass {
    fn bind<'a, A: 'a, B: 'a>(
        fa: Self::F<'a, A>,
        f: impl 'a + FnOnce(A) -> Self::F<'a, B>,
    ) -> Self::F<'a, B> {
        Box::pin(async { f(fa.await).await })
    }

    fn iterate<'a, B: 'a>(mut f: impl Iterative<'a, T = Self, B = B>) -> Self::F<'a, B>
    where
        Self: 'a,
    {
        Box::pin(async move {
            loop {
                match f.next().await {
                    ControlFlow::Continue(next_f) => f = next_f,
                    ControlFlow::Break(b) => return b,
                }
            }
        })
    }

    fn join<'a, A: 'a>(ffa: Self::F<'a, Self::F<'a, A>>) -> Self::F<'a, A> {
        Box::pin(async { ffa.await.await })
    }
}

impl SharedFunctor for FutureClass {
    type Shared<'a, A: 'a + Clone> = Shared<Pin<Box<dyn 'a + Future<Output = A>>>>
    where
        Self: 'a;

    fn share<'a, A: 'a + Clone>(fa: Self::F<'a, A>) -> Self::Shared<'a, A>
    where
        Self: 'a,
    {
        fa.shared()
    }

    fn unshare<'a, A: 'a + Clone>(sa: Self::Shared<'a, A>) -> Self::F<'a, A>
    where
        Self: 'a,
    {
        Box::pin(sa)
    }
}
